%%%-------------------------------------------------------------------
%%% @author Piotr Duleba
%%% @copyright (C) 2020 ACK CYFRONET AGH
%%% This software is released under the MIT license
%%% cited in 'LICENSE.txt'.
%%% @end
%%%-------------------------------------------------------------------
%%% @doc
%%% This module contains functions that create background configuration
%%% and provide access to configuration info.
%%% @end
%%%-------------------------------------------------------------------
-module(oct_background).

-include("oct_background.hrl").
-include_lib("ctool/include/aai/aai.hrl").
-include_lib("ctool/include/aai/caveats.hrl").
-include_lib("ctool/include/test/test_utils.hrl").

%% API
-export([init_per_suite/2]).
-export([
    update_background_config/1,

    to_entity_placeholder/1,
    to_entity_id/1,

    get_provider_id/1,
    get_provider_panels/1,
    get_provider_nodes/1,
    get_provider_eff_users/1,
    get_provider_supported_spaces/1,
    get_provider_ids/0,

    get_zone_panels/0,
    get_zone_nodes/0,

    get_user_id/1,
    get_user_access_token/1,
    get_user_session_id/2,

    get_space_id/1
]).

-type entity_id() :: binary().
-type entity_placeholder() :: atom().
-type onenv_test_config() :: #onenv_test_config{}.

-define(DEFAULT_TEMP_CAVEAT_TTL, 360000).

%%%===================================================================
%%% API
%%%===================================================================


-spec init_per_suite(test_config:config(), onenv_test_config()) ->
    test_config:config().
init_per_suite(Config, #onenv_test_config{
    onenv_scenario = Scenario,
    envs = Envs,
    posthook = CustomPostHook
}) ->
    AddEnvs = lists:map(fun({Service, Application, CustomEnv}) ->
        {add_envs, [Service, Application, CustomEnv]}
    end, Envs),

    test_config:set_many(Config, AddEnvs ++ [
        {set_onenv_scenario, [Scenario]},
        {set_posthook, fun(NewConfig) ->
            CustomPostHook(
                prepare_base_test_config(NewConfig)
            )
        end}
    ]).


-spec update_background_config(test_config:config()) -> test_config:config().
update_background_config(Config) ->
    prepare_base_test_config(Config).


-spec to_entity_placeholder(binary() | atom()) -> entity_placeholder().
to_entity_placeholder(Placeholder) when is_atom(Placeholder) ->
    Placeholder;
to_entity_placeholder(EntityId) when is_binary(EntityId) ->
    kv_utils:get([id_to_placeholder, EntityId], get_oct_mappings()).


-spec to_entity_id(binary() | atom()) -> entity_id().
to_entity_id(Placeholder) when is_binary(Placeholder) ->
    Placeholder;
to_entity_id(EntityId) when is_atom(EntityId) ->
    kv_utils:get([placeholder_to_id, EntityId], get_oct_mappings()).


-spec get_provider_id(atom() | binary()) -> entity_id().
get_provider_id(ProviderIdOrPlaceholder) ->
    ProviderPlaceholder = to_entity_placeholder(ProviderIdOrPlaceholder),
    kv_utils:get([oneproviders, ProviderPlaceholder, id], get_oct_mappings()).


-spec get_provider_panels(atom() | binary()) -> [node()].
get_provider_panels(ProviderIdOrPlaceholder) ->
    ProviderPlaceholder = to_entity_placeholder(ProviderIdOrPlaceholder),
    kv_utils:get([oneproviders, ProviderPlaceholder, panels], get_oct_mappings()).


-spec get_provider_nodes(atom() | binary()) -> [node()].
get_provider_nodes(ProviderIdOrPlaceholder) ->
    ProviderPlaceholder = to_entity_placeholder(ProviderIdOrPlaceholder),
    kv_utils:get([oneproviders, ProviderPlaceholder, nodes], get_oct_mappings()).


-spec get_provider_eff_users(atom() | binary()) -> [entity_id()].
get_provider_eff_users(ProviderIdOrPlaceholder) ->
    ProviderPlaceholder = to_entity_placeholder(ProviderIdOrPlaceholder),
    kv_utils:get([oneproviders, ProviderPlaceholder, users], get_oct_mappings()).


-spec get_provider_supported_spaces(atom() | binary()) -> [entity_id()].
get_provider_supported_spaces(ProviderIdOrPlaceholder) ->
    ProviderPlaceholder = to_entity_placeholder(ProviderIdOrPlaceholder),
    kv_utils:get([oneproviders, ProviderPlaceholder, supported_spaces], get_oct_mappings()).


-spec get_provider_ids() -> [entity_id()].
get_provider_ids() ->
    maps:fold(fun(_, ProviderDetails, IdsAcc) ->
        [kv_utils:get([id], ProviderDetails)| IdsAcc]
    end, [], kv_utils:get([oneproviders], get_oct_mappings())).


-spec get_zone_nodes() -> [node()].
get_zone_nodes() ->
    kv_utils:get([onezones, zone, nodes], get_oct_mappings()).


-spec get_zone_panels() -> [node()].
get_zone_panels() ->
    kv_utils:get([onezones, zone, panels], get_oct_mappings()).


-spec get_user_id(atom() | binary()) -> entity_id().
get_user_id(UserIdOrPlaceholder) ->
    UserPlaceholder = to_entity_placeholder(UserIdOrPlaceholder),
    kv_utils:get([users, UserPlaceholder, id], get_oct_mappings()).


-spec get_user_access_token(atom() | binary()) -> binary().
get_user_access_token(UserIdOrPlaceholder) ->
    UserPlaceholder = to_entity_placeholder(UserIdOrPlaceholder),
    kv_utils:get([users, UserPlaceholder, access_token], get_oct_mappings()).


-spec get_user_session_id(
    UserIdOrPlaceholder :: entity_id() | entity_placeholder(),
    ProviderIdOrPlaceholder :: entity_id() | entity_placeholder()
) ->
    entity_id().
get_user_session_id(UserIdOrPlaceholder, ProviderIdOrPlaceholder) ->
    UserPlaceholder = to_entity_placeholder(UserIdOrPlaceholder),
    ProviderPlaceholder = to_entity_placeholder(ProviderIdOrPlaceholder),
    kv_utils:get([users, UserPlaceholder, sessions, ProviderPlaceholder], get_oct_mappings()).


-spec get_space_id(entity_id() | entity_placeholder()) -> entity_id().
get_space_id(SpaceIdOrPlaceholder) ->
    SpacePlaceholder = to_entity_placeholder(SpaceIdOrPlaceholder),
    kv_utils:get([spaces, SpacePlaceholder, id], get_oct_mappings()).

%%%===================================================================
%%% Internal functions
%%%===================================================================


%% @private
-spec prepare_base_test_config(test_config:config()) -> test_config:config().
prepare_base_test_config(Config) ->
    OneProvidersMap = describe_providers(Config),
    OneZoneMap = describe_onezone(Config),
    UsersMap = describe_users(OneZoneMap, OneProvidersMap),
    SpacesMap = describe_spaces(OneZoneMap),

    {IdToPlaceholderMap, PlaceholderToIdMap} = maps:fold(
        fun(Placeholder, EntityDesc, {IdToPlaceholderAcc, PlaceholderToIdAcc}) ->
            EntityId = maps:get(id, EntityDesc),
            {IdToPlaceholderAcc#{EntityId => Placeholder}, PlaceholderToIdAcc#{Placeholder => EntityId}}
        end,
        {#{}, #{}},
        maps_utils:merge([OneProvidersMap, OneZoneMap, UsersMap, SpacesMap])
    ),

    store_oct_mappings(#{
        oneproviders => OneProvidersMap,
        onezones => OneZoneMap,
        spaces => SpacesMap,
        users => UsersMap,
        id_to_placeholder => IdToPlaceholderMap,
        placeholder_to_id => PlaceholderToIdMap
    }),

    Config.


%% @private
-spec describe_providers(test_config:config()) -> map().
describe_providers(Config) ->
    OpWorkerNodes = test_config:get_all_op_worker_nodes(Config),
    lists:foldl(fun(OpWorkerNode, ProvidersDescAcc) ->
        ProviderId = rpc:call(OpWorkerNode, oneprovider, get_id, []),
        {ok, ProviderName} = rpc:call(OpWorkerNode, provider_logic, get_name, []),
        Placeholder = binary_to_atom(lists:last(binary:split(ProviderName, <<"-">>, [global])), utf8),
        case kv_utils:find([Placeholder, id], ProvidersDescAcc) of
            {ok, ProviderId} ->
                kv_utils:update_with([Placeholder, nodes], fun(Nodes) ->
                    [OpWorkerNode | Nodes]
                end, ProvidersDescAcc);
            error ->
                {ok, Storages} = rpc:call(OpWorkerNode, provider_logic, get_storage_ids, []),
                {ok, Users} = rpc:call(OpWorkerNode, provider_logic, get_eff_users, []),
                Panels = maps:get(ProviderId, get_op_panels_with_id(Config)),
                {ok, Spaces} = rpc:call(OpWorkerNode, provider_logic, get_spaces, []),
                ProviderDetails = #{
                    id => ProviderId,
                    nodes => [OpWorkerNode],
                    panels => Panels,
                    supported_spaces => Spaces,
                    storages => Storages,
                    users => Users
                },
                ProvidersDescAcc#{Placeholder => ProviderDetails};
            {ok, _OtherProviderId} ->
                ct:fail("Wrong enviroment setup. There are multiple providers with the same name: ~p.", [ProviderName])
        end
    end, #{}, OpWorkerNodes).


%% @private
-spec get_op_panels_with_id(test_config:config()) -> map().
get_op_panels_with_id(Config) ->
    lists:foldl(fun(Node, Acc) ->
        [WorkerNode | _] = rpc:call(Node, service_op_worker, get_nodes, []),
        Id = rpc:call(WorkerNode, oneprovider, get_id, []),
        OtherNodes = maps:get(Id, Acc, []),
        Acc#{Id => [Node | OtherNodes]}
    end, #{}, test_config:get_custom(Config, op_panel_nodes, [])).


%% @private
-spec describe_onezone(test_config:config()) -> map().
describe_onezone(Config) ->
    ZoneDetails = #{
        id => <<"onezone">>,
        nodes => test_config:get_all_oz_worker_nodes(Config),
        panels => test_config:get_all_oz_panel_nodes(Config)
    },
    #{zone => ZoneDetails}.


%% @private
-spec describe_users(map(), map()) -> map().
describe_users(OnezoneMap, OneprovidersMap) ->
    [OzNode] = kv_utils:get([zone, nodes], OnezoneMap),
    {ok, UserIds} = rpc:call(OzNode, user_logic, list, [aai:root_auth()]),

    lists:foldl(fun(UserId, Acc) ->
        {ok, UserProtectedData} = rpc:call(OzNode, user_logic, get_protected_data, [aai:root_auth(), UserId]),
        UserName = maps:get(<<"username">>, UserProtectedData),
        Placeholder = binary_to_atom(UserName, utf8),
        AccessToken = create_oz_temp_access_token(OzNode, UserId),

        Sessions = maps:fold(fun(ProviderPlaceholder, ProviderDesc, SessionAcc) ->
            case lists:member(UserId, kv_utils:get([users], ProviderDesc)) of
                true ->
                    [ProviderNode | _] = kv_utils:get([nodes], ProviderDesc),
                    SessionAcc#{ProviderPlaceholder => create_session(ProviderNode, UserId, AccessToken)};
                false ->
                    SessionAcc
            end
        end, #{}, OneprovidersMap),
        Acc#{Placeholder => #{
            id => UserId,
            name => UserName,
            access_token => AccessToken,
            sessions => Sessions
        }}
    end, #{}, UserIds).


%% @private
-spec describe_spaces(map()) -> map().
describe_spaces(OnezoneMap) ->
    [OzNode] = kv_utils:get([zone, nodes], OnezoneMap),
    {ok, SpaceIds} = rpc:call(OzNode, space_logic, list, [aai:root_auth()]),

    lists:foldl(fun(SpaceId, SpaceDescAcc) ->

        {ok, SpaceInfo} = rpc:call(OzNode, space_logic, get_protected_data, [aai:root_auth(), SpaceId]),
        SpaceName = maps:get(<<"name">>, SpaceInfo),
        SpacePlaceholder = binary_to_atom(SpaceName, utf8),

        SpaceDescAcc#{SpacePlaceholder => #{
            id => SpaceId,
            name => SpaceName
        }}

    end, #{}, SpaceIds).


%% @private
-spec get_oct_mappings() -> map().
get_oct_mappings() ->
    node_cache:get(oct_mapping).


%% @private
-spec store_oct_mappings(map()) -> ok.
store_oct_mappings(OctMapping) ->
    node_cache:put(oct_mapping, OctMapping).


%% @private
-spec create_oz_temp_access_token(node(), UserId :: binary()) -> tokens:serialized().
create_oz_temp_access_token(OzNode, UserId) ->
    Auth = ?USER(UserId),
    Now = rpc:call(OzNode, global_clock, timestamp_seconds, []),

    {ok, Token} = rpc:call(OzNode, token_logic, create_user_temporary_token, [Auth, UserId, #{
        <<"type">> => ?ACCESS_TOKEN,
        <<"caveats">> => [#cv_time{valid_until = Now + ?DEFAULT_TEMP_CAVEAT_TTL}]
    }]),
    {ok, SerializedToken} = tokens:serialize(Token),
    SerializedToken.


%% @private
-spec create_session(node(), entity_id(), tokens:serialized()) -> entity_id().
create_session(Node, UserId, AccessToken) ->
    Nonce = crypto:strong_rand_bytes(10),
    Identity = ?SUB(user, UserId),
    TokenCredentials = rpc:call(Node, auth_manager, build_token_credentials, [
        AccessToken, undefined, local_ip_v4(), oneclient, allow_data_access_caveats
    ]),
    {ok, SessionId} = rpc:call(Node, session_manager, reuse_or_create_fuse_session, [Nonce, Identity, TokenCredentials]),
    SessionId.


%% @private
-spec local_ip_v4() -> inet:ip_address().
local_ip_v4() ->
    {ok, Addrs} = inet:getifaddrs(),
    hd([
        Addr || {_, Opts} <- Addrs, {addr, Addr} <- Opts,
        size(Addr) == 4, Addr =/= {127, 0, 0, 1}
    ]).
